use super::{event_loop::EventLoopTask, MqttInterface};
use actix::{Actor, ActorContext, ActorFuture, AsyncContext, Handler, WrapFuture};
use tokio::sync::oneshot;
use tracing::{debug, error, info, trace, warn};

pub(super) struct MqttActor {
    interface: MqttInterface,
    event_loop: Option<rumqttc::EventLoop>,
    quit_rx: Option<oneshot::Receiver<()>>,
}

impl MqttActor {
    pub(super) fn new(interface: MqttInterface, event_loop: rumqttc::EventLoop) -> Self {
        Self {
            interface,
            event_loop: Some(event_loop),
            quit_rx: None,
        }
    }
}

impl Actor for MqttActor {
    type Context = actix::Context<Self>;

    fn started(&mut self, ctx: &mut Self::Context) {
        trace!("MqttActor starting");
        let event_loop = self.event_loop.take().expect("MqttActor already started");
        let (quit_tx, quit_rx) = oneshot::channel();
        ctx.spawn(
            EventLoopTask::new(event_loop, ctx.address(), quit_tx)
                .run()
                .into_actor(self),
        );
        self.quit_rx = Some(quit_rx);
    }
}

#[derive(actix::Message)]
#[rtype(result = "()")]
pub(super) struct Quit;

impl Handler<Quit> for MqttActor {
    type Result = actix::ResponseActFuture<Self, ()>;
    fn handle(&mut self, _msg: Quit, _ctx: &mut Self::Context) -> Self::Result {
        let mut interface = self.interface.clone();
        let quit_rx = self.quit_rx.take().expect("MqttActor has already quit!");
        let fut = async move {
            interface
                .cancel()
                .await
                .expect("could not cancel MQTT client");
            let _ = quit_rx.await;
        }
        .into_actor(self)
        .map(|_, _, ctx| ctx.stop());
        Box::pin(fut)
    }
}

#[derive(actix::Message)]
#[rtype(result = "()")]
pub(super) struct Connected;

impl Handler<Connected> for MqttActor {
    type Result = ();

    fn handle(&mut self, _msg: Connected, ctx: &mut Self::Context) -> Self::Result {
        info!("MQTT connected");
        let mut interface = self.interface.clone();
        let fut = async move {
            let res = interface.publish_connected(true).await;
            let res = res.and(interface.subscribe_requests().await);
            if let Err(err) = res {
                error!("error in connection setup: {}", err);
            }
        };
        ctx.spawn(fut.into_actor(self));
    }
}

#[derive(actix::Message)]
#[rtype(result = "()")]
pub(super) struct PubRecieve(pub(super) rumqttc::Publish);

impl Handler<PubRecieve> for MqttActor {
    type Result = ();

    fn handle(&mut self, msg: PubRecieve, _ctx: &mut Self::Context) -> Self::Result {
        let topic = &msg.0.topic;
        if topic == &self.interface.topics.requests() {
            debug!("received request: {:?}", msg.0);
        } else {
            warn!("received on unknown topic: {:?}", topic);
        }
    }
}
